from somajo import tokenizer
from somajo import sentence_splitter
from somajo import somajo

from .version import __version__

Tokenizer = tokenizer.Tokenizer
SentenceSplitter = sentence_splitter.SentenceSplitter
SoMaJo = somajo.SoMaJo
